from __future__ import annotations

import copy
import importlib
import inspect
import re
from abc import abstractmethod

import numpy as np
import polars

from damast.core.dataframe import AnnotatedDataFrame
from damast.core.metadata import DataSpecification

from .constants import (
    DAMAST_DEFAULT_DATASOURCE,
    DECORATED_DESCRIPTION,
    DECORATED_INPUT_SPECS,
    DECORATED_OUTPUT_SPECS,
    )
from .formatting import DEFAULT_INDENT


class Transformer:
    uuid: str

    def set_uuid(self, uuid: str):
        self.uuid = uuid

    def fit(self, df: AnnotatedDataFrame, other: AnnotatedDataFrame | None = None):
        pass

    def transform(self, df: AnnotatedDataFrame, other: AnnotatedDataFrame | None = None):
        return df

    def fit_transform(self, df: AnnotatedDataFrame, other: AnnotatedDataFrame | None = None):
        if not other:
            self.fit(df=df)
            return self.transform(df=df)
        else:
            self.fit(df, other)
            return self.transform(df,other)

class PipelineElement(Transformer):
    #: Pipeline in which context this processor will be run
    parent_pipeline: 'DataProcessingPipeline'

    #: Map names of input and outputs for a particular pipeline
    _name_mappings: dict[str, dict[str, str]]

    #: Map names of datasource (arguments) to a specific (extra) transformer arguments
    def set_parent(self, pipeline: 'DataProcessingPipeline'):
        """
        Sets the parent pipeline for this pipeline element

        :param pipeline: Parent pipeline
        """
        self.parent_pipeline = pipeline

    @property
    def name_mappings(self) -> dict[str, dict[str, str]]:
        """
        Get current name mappings for this instance
        """
        if not hasattr(self, "_name_mappings"):
            self._name_mappings = { DAMAST_DEFAULT_DATASOURCE: {}}
        return self._name_mappings

    @property
    def parameters(self) -> dict[str, str]:
        """
        Get the current list of parameter for initialization
        """
        if not hasattr(self, "_parameters"):
            self._parameters = None

        self.prepare_parameters()

        return self._parameters

    def prepare_parameters(self):
        parameters = {}
        for p in inspect.signature(self.__init__).parameters:
            parameter_name = p
            if parameter_name not in ['args', 'kwargs']:
                # only process named parameters
                try:
                    parameters[parameter_name] = getattr(self, parameter_name)
                except AttributeError:
                    raise ValueError(f"PipelineElement: please ensure that {self.__class__.__name__}.__init__ keyword arguments"
                            f" are saved in an attribute of the same name: missing '{parameter_name}'")

        self._parameters = parameters

    def get_name(self, name: str, datasource: str | None = None) -> any:
        if datasource is None:
            datasource = DAMAST_DEFAULT_DATASOURCE

        return self._get_name(name=name, datasource=datasource)

    def _get_name(self, name: str, datasource: str | None) -> any:
        """
        Add the fully resolved input/output name for this key.

        :param name: Name as used in the input spec, or pattern "{{x}}_suffix" in order to create a dynamic
                     output based an existing and renameable input
        :param datasource: In cases of multiple input for a node, define the datasource that shall be used
        :return: Name for this input after resolving name mappings and references
        """
        if not isinstance(name, str):
            raise TypeError(f"{self.__class__.__name__}.get_name: provided transformer label is not a string: {name}")

        if datasource is not None:
            try:
                name_mappings = self.name_mappings[datasource]
            except KeyError:
                raise RuntimeError(f"PipelineElement._get_name: not {datasource} in mappings: {self.name_mappings}")
        else:
            name_mappings = self.name_mappings

        if name in name_mappings:
            # allow multiple levels of name resolution, e.g.,
            # x -> y, y -> z --> x -> z
            mapped_name = name_mappings[name]
            if mapped_name == name:
                return name

            return self._get_name(mapped_name, datasource=datasource)

        # Allow to use patterns, so that an existing input
        # reference can be reused for dynamic labelling
        while re.search("{{\\w+}}", name):
            for match in re.finditer("{{\\w+}}", name):
                resolved_name = match.group()[2:-2]
                if resolved_name in name_mappings:
                    resolved_name = name_mappings[resolved_name]

                name = name.replace(match.group(), resolved_name)
        return name

    @abstractmethod
    def transform(self, df: AnnotatedDataFrame) -> AnnotatedDataFrame:
        """
        Default transform implementation
        """

    @property
    def input_specs(self) -> dict[str, list[DataSpecification]]:
        if not hasattr(self.transform, DECORATED_INPUT_SPECS):
            raise AttributeError(
                f"{self.__class__.__name__}.validate: missing input specification"
                f" for processing step '{self.__class__.__name__}'"
            )

        generic_spec = getattr(self.transform, DECORATED_INPUT_SPECS)
        specs = copy.deepcopy(generic_spec)
        for label, speclist in specs.items():
            for spec in speclist:
                spec.name = self.get_name(spec.name, label)

        return specs

    @property
    def output_specs(self) -> list[DataSpecification]:
        if not hasattr(self.transform, DECORATED_OUTPUT_SPECS):
            raise AttributeError(
                f"{self.__class__.__name__}.validate: missing output specification"
                f" for processing step '{self.__class__.__name__}'"
            )

        generic_spec = getattr(self.transform, DECORATED_OUTPUT_SPECS)
        specs = copy.deepcopy(generic_spec)
        for spec in specs:
            # there will be only 1 dataframe as output
            spec.name = self.get_name(spec.name)

        return specs

    @classmethod
    def create_new(cls,
                   module_name: str,
                   class_name: str,
                   name_mappings: dict[str, dict[str, str]] | None = None,
                   parameters: dict[str, any] | None = {}) -> PipelineElement:
        """
        Create a new PipelineElement Subclass instance dynamically

        :param module_name: Name of the module for the PipelineElement class
        :param class_name: Name of the PipelineElement subclass
        :param name_mappings: Dictionary of name mappings that should apply
        :return: Instance for the PipelineElement instance

        :raise ValueError: If module or class with given name is not specified
        :raise ImportError: If class could not be loaded
        """
        if module_name is None:
            raise ValueError(f"{cls.__name__}.create_new: missing 'module_name'")

        if class_name is None:
            raise ValueError(f"{cls.__name__}.create_new: missing 'class_name'")

        p_module = importlib.import_module(module_name)
        if hasattr(p_module, class_name):
            klass = getattr(p_module, class_name)
        else:
            raise ImportError(f"{cls.__name__}.create_new: could not load '{class_name}' from '{p_module}'")
        if parameters:
            instance = klass(**parameters)
        else:
            instance = klass()

        if name_mappings is None:
            name_mappings = { DAMAST_DEFAULT_DATASOURCE: {}}

        instance._name_mappings = name_mappings
        return instance

    def __iter__(self):
        yield "module_name", f"{self.__class__.__module__}"
        yield "class_name", f"{self.__class__.__qualname__}"
        yield "parameters", self.parameters
        yield "name_mappings", self.name_mappings

    def __eq__(self, other):
        return dict(self) == dict(other)

    @classmethod
    def get_types(cls) -> list[PipelineElement]:
        """
        Get all available PipelineElement implementations

        :return: List of PipelineElement classes
        """
        klasses = []
        for subclass in cls.__subclasses__():
            klasses.append(subclass)
            klasses.extend(subclass._subclasses())
        return klasses

    @classmethod
    def _subclasses(cls) -> list[PipelineElement]:
        """
        Generate the list of subclasses for the calling class
        """
        klasses = []
        for subclass in cls.__subclasses__():
            klasses.append(subclass)
            klasses.extend(subclass._subclasses())
        return klasses

    @classmethod
    def generate_subclass_documentation(cls) -> str:
        """
        Generate the documentation for all subclasses of ::class::`PipelineElement`
        """
        implementations = sorted(cls.get_types(), key=str)
        txt = ""
        for k in implementations:
            txt += "=" * 80
            txt += f"\n{k.__name__} -- from {k.__module__}\n"
            txt += k.__doc__
            txt += '\n'
        return txt

    def to_str(self, indent_level: int = 0) -> str:
        hspace = DEFAULT_INDENT * indent_level
        data = hspace + self.__class__.__name__ + "\n"
        if hasattr(self.transform, DECORATED_DESCRIPTION):
             description = getattr(self.transform, DECORATED_DESCRIPTION)
             data += (
                     hspace + DEFAULT_INDENT * 2 + "description: " + description + "\n"
                )

        data += hspace + DEFAULT_INDENT * 2 + "input:\n"
        if hasattr(self.transform, DECORATED_INPUT_SPECS):
            data += DataSpecification.to_str(
                self.input_specs, indent_level=indent_level + 4
            )

        data += hspace + DEFAULT_INDENT * 2 + "output:\n"
        if hasattr(self.transform, DECORATED_OUTPUT_SPECS):
            data += DataSpecification.to_str(
                self.output_specs, indent_level=indent_level + 4
            )
        return data

    def __deepcopy__(self, memo) -> PipelineElement:
        new_element = copy.copy(self)
        new_element._name_mappings = copy.deepcopy(self.name_mappings)
        return new_element

# Only for internal use
class MultiCycleTransformer(Transformer):
    def __init__(self, features: list[str], n: int):
        self.features = features
        self.n = n

    def transform(self, df: AnnotatedDataFrame):
        if type(df) is not AnnotatedDataFrame:
            raise ValueError(f"Transformer requires 'AnnotatedDataFrame',"
                    f" but got '{type(df)}")
        clone = df.copy()

        for feature in self.features:
            clone._dataframe = clone._dataframe.with_columns(
                    (np.cos(polars.col(feature)*2*np.pi) / self.n).alias(f"{feature}_x")
                )
            clone._dataframe = clone._dataframe.with_columns(
                    (np.cos(polars.col(feature)*2*np.pi) / self.n).alias(f"{feature}_y")
                )
        return clone




from __future__ import annotations

import copy
import inspect
import uuid as uuid_module

import networkx as nx

from damast.core.dataframe import AnnotatedDataFrame
from damast.core.formatting import DEFAULT_INDENT
from damast.core.metadata import DataSpecification
from damast.core.transformations import PipelineElement

from .decorators import DAMAST_DEFAULT_DATASOURCE, describe, input, output


class DataSource(PipelineElement):
    """
    PipelineElement that marks a datasource, e.g.,
    an annotated dataframe that needs to be processed
    """
    def __init__(self):
        super().__init__()

    @describe("Node for marking a plain data entry")
    @input({})
    @output({})
    def transform(self, df: AnnotatedDataFrame) -> AnnotatedDataFrame:
        return df


class Node:
    name: str
    transformer: PipelineElement

    #: Cache computation results
    result: AnnotatedDataFrame
    #: Cache validation results
    validation_output_spec: list[DataSpecification]

    def __init__(self, name: str, transformer: PipelineElement, uuid: str | None = None):
        self.name = name
        self.transformer = transformer

        if uuid is None:
            self.transformer.set_uuid(str(uuid_module.uuid4()))
        else:
            self.transformer.set_uuid(uuid)

    def __repr__(self):
        return f"name={self.name} {self.transformer.__class__.__name__} (uuid={self.uuid})"

    def inputs(self) -> list[str]:
        """
        Get all inputs for AnnotatedDataFrame
        """
        inputs = []
        for x, y in inspect.signature(self.transformer.transform).parameters.items():
            if y.annotation == 'AnnotatedDataFrame' or y.annotation.__name__ == 'AnnotatedDataFrame':
                inputs.append(x)
        return inputs

    def is_datasource(self) -> bool:
        return type(self.transformer) is DataSource

    @property
    def uuid(self):
        return self.transformer.uuid

    def to_str(self, indent_level: int = 0) -> str:
        """
        Output as string.

        :param indent_level: Indentation per step.
            It is multiplied by :attr:`damast.core.formatting.DEFAULT_INDENT`.
        """
        hspace = DEFAULT_INDENT * indent_level

        data = hspace + DEFAULT_INDENT + self.name + ":\n"
        data += self.transformer.to_str(indent_level+4)
        return data

    def __iter__(self):
        yield "uuid", str(self.uuid)
        yield "name", self.name
        yield "transformer", dict(self.transformer)

    @classmethod
    def from_dict(cls, data: dict[str, any]):
        return Node(uuid=data['uuid'],
                    name=data['name'],
                    transformer=PipelineElement.create_new(**data['transformer'])
                )


class ProcessingGraph:
    _graph: nx.DiGraph

    _root_node: Node
    _leaf_node: Node

    def __init__(self, with_datasource: bool = True):
        self._graph = nx.DiGraph()
        self._root_node = None
        self._leaf_node = None

        if with_datasource:
            self.add_datasource()

    def add_datasource(self):
        """
        Add a datasource node as root node to the graph
        (only if the root node has not been set)
        """
        if self._root_node:
            raise RuntimeError("ProcessingGraph.add_datasource: adding a "
                               " datasource is only possible for an empty graph")

        self._root_node = Node(DAMAST_DEFAULT_DATASOURCE, DataSource())
        self._graph.add_node(self._root_node)

        self._leaf_node = self._root_node

    def __getitem__(self, uuid: str):
        for n in self._graph.nodes():
            if n.uuid == uuid:
                return n

        raise KeyError(f"Could not find node with {uuid=}")

    def __iter__(self):
        yield "nodes", [dict(x) for x in self._graph.nodes()]
        yield "edges", [{'from': str(x[0].uuid), 'to': str(x[1].uuid)} for x in self._graph.edges()]

    @classmethod
    def from_dict(cls, data: dict[str, any]):
        """
        Load the graph from dictionary containing nodes and edges.
        Note, that no internal states can be loaded.
        """
        if 'nodes' not in data:
            raise KeyError("ProcessingGraph.from_dict: missing 'nodes'")

        if 'edges' not in data:
            raise KeyError("ProcessingGraph.from_dict: missing 'edges'")

        graph = cls(with_datasource=False)

        for node_dict in data['nodes']:
            node = Node.from_dict(node_dict)
            graph.add(node)

        for edge_dict in data['edges']:
            from_node = graph[edge_dict['from']]
            to_node = graph[edge_dict['to']]

            graph.add_edge(from_node, to_node)

        return graph

    def __eq__(self, other: ProcessingGraph) -> bool:
        """
        Compares the underlying graphs
        """
        return nx.utils.graphs_equal(self._graph, other._graph)

    def add(self, node: Node):
        """
        Add (or rather appaned) a node to the current graph
        It will attach to the graph's current leaf node
        """
        if not self._root_node:
            self._root_node = node

        self._graph.add_node(node)

        required_slots = node.inputs()
        if len(required_slots) != 1:
            raise RuntimeError(f"{node} needs to have exactly one dataframe input argument",
                               " found {required_slots}")

        if self._leaf_node is not None:
            self._graph.add_edge(self._leaf_node, node, slot=required_slots[0])
        self._leaf_node = node

    def join(self, name: str, operator: PipelineElement, processing_graph: ProcessingGraph | None = None):
        """
        Join another processing graph into the current one.
        This will an new 'join' node for the given operator.
        If no processing_graph is given, then a DataSource node will be created and added to the
        graph label by :name
        """
        leaf_node = None

        # if this is to join a processing graph
        # add all nodes and edges
        if processing_graph:
            for n in processing_graph.nodes():
                # Reset the source node of this processing graph
                # to the name of this join
                if n.is_datasource():
                    n.name = name

                self._graph.add_node(n)
                if self._graph.out_degree(n) == 0:
                    leaf_node = n

            for from_n, to_n, data in processing_graph.edges(data=True):
                self._graph.add_edge(from_n, to_n, **data)

        if not leaf_node:
            # we require at least a data loader stub node
            # by convention name after the 'join' node
            leaf_node = Node(name, DataSource())
            self._graph.add_node(leaf_node)

        # we consider order here and assume max 2
        node = Node(name, operator)
        required_slots = node.inputs()
        if len(required_slots) != 2:
            raise RuntimeError(f"ProcessingGraph.join: requires exactly two dataframe argument, but found {required_slots}")

        self._graph.add_node(node)

        # Allow to point to the argument in the transformer / operator which will
        # receive the input from the connection predecessor
        self._graph.add_edge(leaf_node, node, slot=required_slots[1])
        self._graph.add_edge(self._leaf_node, node, slot=required_slots[0])

        self._leaf_node = node

    def to_str(self, indent_level = 0):
        return '\n'.join(list(nx.generate_network_text(self._graph, ascii_only=True)))
        #data = ""
        #nx.write_network_text(self._graph)
        #reversed_graph = self._graph.reverse()
        #for node in nx.bfs_tree(reversed_graph, self._leaf_node):
        #    if len(list(self._graph.successors(node))) == 2:
        #        print("JOIN NODE")
        #    data += node.to_str(indent_level=indent_level)
        ##nx.write_network_text(reversed_graph)
        #return data

    def get_joins(self, in_degree: int = 2):
        """
        Get nodes, that have an in_degree of :degree (default is two)
        """
        if in_degree < 2:
            raise ValueError("ProcessingGraph.get_joins: no join nodes with an in degree"
                             " of less than 2 possible")
        joins = []
        for n in self._graph.nodes():
            if self._graph.in_degree(n) == in_degree:
                joins.append(n)
        return joins

    def __getattr__(self, name):
        # forward calls to the underlying graph
        return getattr(self._graph, name)

    def inputs_ready(self, node: Node) -> bool:
        """
        Check if the results from predecessor nodes that this
        node computation depends upon are ready for ingestion
        """
        for i in node.inputs():
            for x,y in self._graph.in_edges(node):
                if not hasattr(x, 'result') or x.result is None:
                    return False

        return True

    def get_current_inputs(self, node: Node) -> bool:
        """
        Get the current inputs
        """
        inputs = {}
        for i in node.inputs():
            for x,y,data in self._graph.in_edges(node, data=True):
                if hasattr(x, 'result'):
                    inputs[data['slot']] = x.result
        return inputs

    def clear_state(self):
        """
        Clear the internal state including computation and validation results
        """
        for n in self._graph.nodes():
            n.result = None
            n.validation_output_spec = None

    def __len__(self) -> int:
        return len(self._graph)

    def __deepcopy__(self, memo) -> ProcessingGraph:
        new_graph = ProcessingGraph(with_datasource=False)
        for n in self._graph.nodes():
            new_node = Node(name=f"{n.name}", transformer=copy.deepcopy(n.transformer), uuid=f"{n.uuid}")
            new_graph.add_node(new_node)

        for from_n, to_n, data in self._graph.edges(data=True):
            new_graph.add_edge(new_graph[from_n.uuid], new_graph[to_n.uuid], **data.copy())

        return new_graph


    def execute(self, node: Node) -> AnnotatedDataFrame:
        """
        Run the transformation of this node
        :raise RuntimeError when the results from other nodes are not yet available
        """
        kwargs = {}
        if not self.inputs_ready(node):
            raise RuntimeError(f"ProcessingGraph.node: {node} inputs are not ready")

        for i in node.inputs():
            for from_node,to_node,data in self._graph.in_edges(node, data=True):
                if data['slot'] == i:
                    kwargs[data['slot']] = from_node.result
        return node.transformer.fit_transform(**kwargs)





__all__ = [
    "DECORATED_ARTIFACT_SPECS",
    "DECORATED_DESCRIPTION",
    "DECORATED_INPUT_SPECS",
    "DECORATED_OUTPUT_SPECS",
]
DAMAST_DEFAULT_DATASOURCE = "df"

DECORATED_DESCRIPTION = "_damast_description"
"""Attribute description for :func:`describe`"""

DECORATED_ARTIFACT_SPECS = "_damast_artifact_specs"
"""Attribute description for :func:`artifacts`"""

DECORATED_INPUT_SPECS = "_damast_input_specs"
"""Attribute description for :func:`input`"""

DECORATED_OUTPUT_SPECS = "_damast_output_specs"
"""Attribute description for :func:`output`"""

DAMAST_HDF5_ROOT: str = "/dataframe"
DAMAST_HDF5_COLUMNS: str = "/dataframe/columns"

DAMAST_SPEC_SUFFIX: str = ".spec.yaml"
DAMAST_SUPPORTED_FILE_FORMATS: dict[str, list[str]] = {
    "parquet": [".parquet", ".pq"],
    "netcdf": [".nc", ".netcdf", ".nc4", ".netcdf4"],
    "hdf": [".h5", ".hdf", ".hdf5"],
    "csv": [".csv"],
}

DAMAST_MOUNT_PREFIX: str = "damast-mount"

DAMAST_CSV_DEFAULT_ARGS: dict[str, any] = {
    'null_values': ["None", "none", "Null", "null"],
    'quote_char': None,
    'infer_schema_length': None
}



"""
Module which collects transformers that add / augment the existing data
"""
from pathlib import Path
from typing import Callable, Union

import numpy as np
import numpy.typing as npt
import polars as pl

import damast.core
import damast.data_handling.transformers.augmenters as augmenters
from damast.core.dataprocessing import PipelineElement
from damast.core.types import DataFrame, XDataFrame
from damast.data_handling.transformers.augmenters import BallTreeAugmenter
from damast.domains.maritime.ais.navigational_status import AISNavigationalStatus
from damast.domains.maritime.ais.vessel_types import VesselType
from damast.domains.maritime.math.spatial import EARTH_RADIUS

__all__ = [
    "AddMissingAISStatus",
    "ComputeClosestAnchorage"
]


class ComputeClosestAnchorage(PipelineElement):
    """
    Compute the closest anchorage given a data-set with all closest anchorages

    :param dataset: Path to data-set with closest anchorages
    :param columns: Names of columns used to define the distance to anchorage (The data should be in degrees)
    :param sep: Separator used in dataset if dataset is a csv file
    """
    _function: Callable[[npt.NDArray[np.float64], npt.NDArray[np.float64]], npt.NDArray[np.float64]]

    dataset: str
    columns: list[str]
    sep: str

    @property
    def deg2rad(self) -> float:
        return np.pi/180.0

    def __init__(self,
                 dataset: str | Path,
                 columns: list[str],
                 sep: str = ";"):
        self.dataset = dataset
        self.columns = columns
        self.sep = sep

        if type(dataset) in [str, Path]:
            _dataset = self.load_data(dataset, sep)
        else:
            _dataset = dataset

        radian_dataset = [
                _dataset.select((pl.col(x)*self.deg2rad).alias(x))[:,0] for x in columns
        ]

        self._function = BallTreeAugmenter(np.vstack(radian_dataset).T, "haversine")

    @classmethod
    def load_data(cls,
                  filename: Union[str, Path], sep: str) -> DataFrame:
        """
        Load dataset from file

        :param filename: The input file (or path)
        :param sep: Separator in csv
        :return: A `pandas.DataFrame` where each row has a column MMSI and vessel_type
        """
        try:
            return XDataFrame.open(path=filename, sep=sep)
        except FileNotFoundError:
            raise RuntimeError(f"{cls}: Vessel type information not accessible. File {vessel_type_csv} not found")

    @damast.core.describe("Compute distance from dataset to closest anchorage")
    @damast.core.input({"x": {"representation_type": float, "unit": damast.core.units.units.deg},
                        "y": {"representation_type": float, "unit": damast.core.units.units.deg}})
    @damast.core.output({"distance": {"representation_type": float, "unit": damast.core.units.units.km}})
    def transform(self, df:  damast.core.AnnotatedDataFrame) -> damast.core.AnnotatedDataFrame:
        dataframe = df._dataframe

        x_name = self.get_name('x')
        y_name = self.get_name('y')

        # Transform latitude and longitude to radians
        dataframe = dataframe.with_columns(
            (pl.col(x_name)*self.deg2rad).alias(f"{x_name}_rad")
        )
        dataframe = dataframe.with_columns(
            (pl.col(y_name)*self.deg2rad).alias(f"{y_name}_rad")
        )

        distance = self.get_name('distance')
        dataframe = dataframe.with_columns(
            pl.struct(f"{x_name}_rad", f"{y_name}_rad").map_batches(
                lambda x: self._function(x.struct.field(f"{x_name}_rad"), x.struct.field(f"{y_name}_rad")),
                return_dtype=float
            ).alias(distance)
        )

        # Multiply distance column by earth radius
        dataframe = dataframe.with_columns(
            (pl.col(distance)*EARTH_RADIUS).alias(distance)
        )
        #dataframe.units[self.get_name("distance")] = damast.core.units.units.km

        # Drop/hide conversion columns
        df._dataframe = dataframe.drop([f"{self.get_name('x')}_rad", f"{self.get_name('y')}_rad"])
        return df


class AddMissingAISStatus(augmenters.AddUndefinedValue):
    """For a given column replace rows with missing entries with
    :attr:`damast.domains.maritime.ais.AISNavigationalStatus.Undefined`.
    """

    def __init__(self):
        self._fill_value = int(AISNavigationalStatus.Undefined)

    @damast.core.describe("Fill missing AIS status")
    @damast.core.input({"x": {"representation_type": int}})
    @damast.core.output({"x": {"representation_type": int}})
    def transform(self, df: damast.core.AnnotatedDataFrame) -> damast.core.AnnotatedDataFrame:
        return super().transform(df)


class AddVesselType(augmenters.JoinDataFrameByColumn):
    """
    Add the vessel type (as integer value) based on external data-set and corresponding to ::class::`VesselType`.

    The vessel type might for instance come from the [global fishing watch database](https://globalfishingwatch.org/).
    It must be either a string corresponding to the snake_case class name, or the corresponding integer

    :param right_on: Name in data-set column to use for merging datasets
    :param dataset_column: Column to add to input dataframe
    :param dataset: Dataset or path to dataset
    :raise ValueError: when the input column that shall extend the dataset is neither int nor str
    """

    def __init__(self,
                 right_on: str,
                 dataset_column: str,
                 dataset: Union[str, Path, DataFrame]
                 ):

        if type(dataset) in [str, Path]:
            dataset = XDataFrame.open(path=dataset)

        column_dtype = XDataFrame(dataset).dtype(dataset_column)
        name = f"{dataset_column}_mapped"
        if str in [column_dtype, column_dtype.to_python()]:
            # VesselTypes should be mapped to integers
            mapping = VesselType.get_mapping()
            dataset = dataset.with_columns(
                pl.col(dataset_column).replace_strict(mapping).alias(name)
            )
        elif int in [column_dtype, column_dtype.to_python()]:
            dataset = dataset.with_columns(
                    pl.col(dataset_column).alias(name)
            )
        else:
            raise ValueError(f"{self.__class__.__name__}.__init__: dtype of column '{dataset_column}',"
                             f" must be either int or str, but was '{column_dtype}'")
        super().__init__(dataset=dataset, right_on=right_on, dataset_column=name)

    @damast.core.describe("Add vessel-type from other dataset to current dataset, "
                          "where the input x is the linking identifier")
    @damast.core.input({"x": {}})
    @damast.core.output({"out": {"representation_type": int}})
    def transform(self, df: damast.core.AnnotatedDataFrame) -> damast.core.AnnotatedDataFrame:
        return super().transform(df)



"""
Module containing decorators and classes to model data processing pipelines
"""
from __future__ import annotations

import copy
import importlib
import os
import tempfile
import traceback as tc
from collections import OrderedDict
from datetime import datetime, timezone
from logging import Logger, getLogger
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import yaml
from tqdm import tqdm

import damast.version
from damast.core.processing_graph import Node, ProcessingGraph
from damast.core.transformations import PipelineElement

from .constants import DAMAST_DEFAULT_DATASOURCE
from .dataframe import AnnotatedDataFrame
from .metadata import DataSpecification, MetaData

__all__ = [
    "DataProcessingPipeline",
    "PipelineElement",
]

logger: Logger = getLogger(__name__)

DAMAST_PIPELINE_SUFFIX: str = ".damast.ppl"
"""Suffix of :class:`DataProcessingPipeline` files created in :func:`DataProcessingPipeline.save`
and used by :func:`DataProcessingPipeline.load`"""

VAEX_STATE_SUFFIX: str = ".vaex-state.json"
"""Suffix of :class:`DataProcessingPipeline` files created in :func:`DataProcessingPipeline.save_state`
and used by :func:`DataProcessingPipeline.load_state`"""


class DataProcessingPipeline(PipelineElement):
    """
    A data-processing pipeline for a sequence of transformers

    :param name: Name of the pipeline
    :param base_dir: Base directory towards which transformer output which be relative
    :param steps: Steps that form the basis for this pipeline
    :param inplace_transformation: If true, the input :class:`damast.core.dataframe.AnnotatedDataFrame` is not
        copied when calling :func:`transform`. Else input data-frame is untouched
    :param name_mappings: Name mappings that should be applied to individual transformations.

    :raises ValueError: If any of the transformer names are `None`
    :raises AttributeError: If the transformer is missing the :func:`transform` function
    :raises AttributeError: If transformer is missing input or output decoration
    :raises RuntimeError: If the sequence of transformers does not satisfy the sequential requirements
    """

    #: Name of the processing pipeline
    name: str
    description: str

    #: Base path (which is forwarded to transformers, when calling
    #: transform)
    base_dir: Path

    #: The output specs - as specified by decorators
    _output_specs: List[DataSpecification]

    #: Check if the pipeline is ready to be run
    is_ready: bool
    # The processing graph that define this pipeline
    processing_graph: ProcessingGraph

    _inplace_transformation: bool
    _name_mappings: Dict[str, Dict[str, str]]
    _processing_stats: Dict[str, Dict[str, Any]]

    _meta: Dict[str, str]

    def __init__(self, *,
                 name: str,
                 description: str = "",
                 base_dir: Union[str, Path] = tempfile.gettempdir(),
                 processing_graph: List[Tuple[str, Union[Dict[str, Any], PipelineElement]]] | ProcessingGraph = None,
                 inplace_transformation: bool = False,
                 name_mappings: Dict[str, Dict[str, str]] = { DAMAST_DEFAULT_DATASOURCE: {}},
                 meta: Dict[str, str] | None = None,
                 ):
        super().__init__()

        self.name = name
        self.description = description
        self.base_dir = Path(base_dir)

        self._output_specs = []
        self._inplace_transformation = inplace_transformation

        self._name_mappings = name_mappings
        self._processing_stats = {}

        self.processing_graph = ProcessingGraph()
        if processing_graph:
            if isinstance(processing_graph, ProcessingGraph):
                self.processing_graph = processing_graph
            elif isinstance(processing_graph, dict):
                self.processing_graph = ProcessingGraph.from_dict(processing_graph)
            elif isinstance(processing_graph, list):
                for step in processing_graph:
                    name, instance = step
                    if isinstance(instance, PipelineElement):
                        self.processing_graph.add(Node(
                                name=name,
                                transformer=instance
                            )
                        )
                    elif isinstance(instance, dict):
                        self.processing_graph.add(Node(
                                name=name,
                                transformer=PipelineElement.create_new(**instance)
                            )
                        )
                    else:
                        raise ValueError(f"{self.__class__.__name__}.__init__: could not instantiate PipelineElement"
                                     f" from {type(step)}")
        self.is_ready = False

        if meta is not None:
            self._meta = meta
        else:
            self._meta = {
                'damast_version': damast.version.__version__
            }

    @property
    def output_specs(self):
        if not self.is_ready:
            raise RuntimeError(
                f"{self.__class__.__name__}.output_specs: pipeline is not yet ready to run. "
                f"Please call 'prepare()' to set the correct output specs"
            )

        return self._output_specs

    def add(
            self,
            name: str,
            transformer: PipelineElement,
            name_mappings: Optional[Dict[str, str]] = None,
    ) -> DataProcessingPipeline:
        """
        Add a pipeline step

        :param name: Name of the step
        :param transformer: The transformer that shall be executed
        :param name_mappings: Allow to define a name mapping for this pipeline element instance
        :return:
        """
        transformer.set_parent(pipeline=self)
        if name_mappings is not None:
            if len(transformer.input_specs) == 1 and DAMAST_DEFAULT_DATASOURCE not in name_mappings:
                transformer._name_mappings = { DAMAST_DEFAULT_DATASOURCE: name_mappings.copy() }

        self.processing_graph.add(
                Node(name=name,
                     transformer=transformer
                     )
        )
        self.is_ready = False
        return self

    def join(
            self,
            name: str,
            operator: PipelineElement,
            data_source: DataProcessingPipeline | None = None,
            name_mappings: Optional[Dict[str, str]] = None,
    ) -> DataProcessingPipeline:
        """
        Add a pipeline step

        :param name: Name of the step
        :param transformer: The transformer that shall be executed
        :param name_mappings: Allow to define a name mapping for this pipeline element instance
        :return:
        """
        operator.set_parent(pipeline=self)
        if name_mappings is not None:
            operator._name_mappings = name_mappings.copy()

        self.processing_graph.join(
                name=name,
                processing_graph=data_source.processing_graph if data_source else None,
                operator=operator
        )
        self.is_ready = False
        return self

    @classmethod
    def validate(
            cls, processing_graph: ProcessingGraph, metadata: dict[str, MetaData]
    ) -> Dict[str, Any]:
        """
        Validate the existing pipeline and collect the minimal input and output data specification.

        :param steps: processing steps
        :param metadata: the input metadata for this pipeline as dictionary of input datasource to metadata
        :return: The minimal output specification for this pipeline
        """
        processing_graph.clear_state()

        current_specs: dict[list[DataSpecification]] = {}
        for ds_name, ds_metadata in metadata.items():
            # Keep track of the expected (minimal) specs at each step in the pipeline
            current_specs[ds_name] = copy.deepcopy(
                ds_metadata.columns
            )

        for idx, node in enumerate(processing_graph.nodes(), start=1):
            if node.name is None:
                raise ValueError(
                    f"{cls.__name__}.validate: missing name processing step"
                )

            if not hasattr(node.transformer, "transform"):
                raise AttributeError(
                    f"{cls.__name__}.validate: processing step '{node.name}' does not fulfill the"
                    f" TransformerMixin requirements - no method 'fit_transform' found"
                )

            try:
                logger.info("#{idx} validate {node}")
                if node.name in current_specs and node.is_datasource():
                    datasource = node.name
                    md = MetaData(columns=current_specs[datasource], annotations=[])
                    node_input_specs = list(node.transformer.input_specs.values())[0]
                    fulfillment = md.get_fulfillment(expected_specs=node_input_specs)
                    if not fulfillment.is_met():
                        raise RuntimeError(
                            f"{cls.__name__}.validate: Input requirements are not fulfilled (for datasource '{node.transformer}'). "
                            f"Current input available (at step '{node.name}'): {fulfillment}"
                        )
                    node.validation_output_spec = DataSpecification.merge_lists(node_input_specs,
                                                                                current_specs[datasource])
                else:
                    for i in node.inputs():
                        for from_node,to_node,data in processing_graph._graph.in_edges(node, data=True):
                            slot = data['slot']

                            md = MetaData(columns=from_node.validation_output_spec, annotations=[])
                            fulfillment = md.get_fulfillment(expected_specs=node.transformer.input_specs[slot])
                            if not fulfillment.is_met():
                                raise RuntimeError(
                                    f"{cls.__name__}.validate: Input requirements are not fulfilled {node} (for {slot=}). "
                                     f"Current input available (at step '{node.name}'): {fulfillment}"
                                )

                            if not node.validation_output_spec:
                                node.validation_output_spec = DataSpecification.merge_lists(from_node.validation_output_spec, node.transformer.output_specs)
                            else:
                                node.validation_output_spec = DataSpecification.merge_lists(node.transformer.output_specs, node.validation_output_spec)

            except Exception as e:
                msg = ''.join(tc.format_exception(e)[-2:])
                raise RuntimeError(f"Validation of step #{idx} in pipeline ({node}) failed: name_mappings: {node.transformer.name_mappings}"
                                   f"{msg}")

        return {"processing_graph": processing_graph, "output_spec": node.validation_output_spec}

    def save(self, dir: Union[str, Path]) -> Path:
        """
        Save the processing pipeline

        :param dir: directory where to save this pipeline
        """
        base_dir = Path(dir)
        base_dir.mkdir(parents=True, exist_ok=True)
        filename = base_dir / f"{self.name}{DAMAST_PIPELINE_SUFFIX}"

        with open(filename, "w") as f:
            yaml.dump(dict(self), f)
        return filename

    def __iter__(self):
        yield "name", self.name
        yield "description", self.description
        yield "meta", self._meta
        yield "base_dir", str(self.base_dir)
        yield "processing_graph", dict(self.processing_graph)

    def save_state(self,
                   df: AnnotatedDataFrame,
                   dir: Union[str, Path]) -> Path:
        """
        Save the processing pipeline

        :param dir: directory where to save this pipeline
        """
        filename = Path(dir) / f"{self.name}{VAEX_STATE_SUFFIX}"
        #df._dataframe.state_write(file=filename)
        df._dataframe.serialize(filename)
        return filename

    @classmethod
    def load(cls, path: Union[str, Path], name: str = "*") -> DataProcessingPipeline:
        """
        Load a :class:`DataProcessingPipeline` from file (without suffix :attr:`DAMAST_PIPELINE_SUFFIX`)

        :param path: Directory containing pipeline(s) (with fiven suffix) or pipeline filename
        :name: Name of file(s) without suffix. Can be a Regex expression
        """
        basename = f"{name}{DAMAST_PIPELINE_SUFFIX}"
        path = Path(path)
        if path.is_dir():
            files = list(path.glob(basename))
        elif path.is_file():
            files = [path]
        else:
            raise RuntimeError("{self.__class__.__name__}.load: {path}"
                               " is neither directory nor file")

        if len(files) == 1:
            filename = files[0]
        elif len(files) == 0:
            raise FileNotFoundError(
                f"{cls.__name__}.load: could not find '{basename}' in {dir}"
            )
        else:
            raise RuntimeError(
                f"{cls.__name__}.load: multiple pipeline available, pls be more specific:"
                f" {','.join([x.name for x in files])}"
            )

        with open(filename, "r") as f:
            data = yaml.load(f, Loader=yaml.SafeLoader)

        return cls(**data)

    @classmethod
    def load_state(
            cls, df: AnnotatedDataFrame, dir: Union[str, Path], name: str = "*"
    ) -> AnnotatedDataFrame:
        """
        Load a ``vaex`` state (from file) to a :class:`damast.core.AnnotatedDataFrame`.

        :param df: The data-frame
        :param dir: Directory of vaex state file
        :param name: Regex describing vaex state file (without :attr:`VAEX_STATE_SUFFIX`)
        :raises FileNotFoundError: If no files found
        :raises RuntimeError: If multiple vaex-states satisfies the ``name`` requirement.
        :return: The data-frame with applied state
        """
        basename = f"{name}{VAEX_STATE_SUFFIX}"
        dir_path = Path(dir)

        files = list(dir_path.glob(basename))
        if len(files) == 1:
            filename = files[0]
        elif len(files) == 0:
            raise FileNotFoundError(
                f"{cls.__name__}.load: could not find '{basename}' in {dir_path}"
            )
        else:
            raise RuntimeError(
                f"{cls.__name__}.load: multiple pipeline available, pls be more specific:"
                f" {','.join([x.name for x in files])}"
            )

        #df.dataframe.state_load(file=filename)
        df._dataframe = df.dataframe.deserialize(filename)
        return df

    def prepare(self,
                name_mappings: dict[str, dict[str, any]] | None = None,
                **dataframes) -> DataProcessingPipeline:
        """
        Prepare the pipeline by applying necessary name mapping and validating all steps that define the pipeline.

        A :class:`DataProcessingPipeline` must be prepared before execution.

        :returns: This instance of data processing pipeline
        """
        if not dataframes:
            raise RuntimeError("DataProcessingPipeline.prepare: missing dataframes")

        pipeline = copy.deepcopy(self)

        # Update the pipeline mapping with a temporary overlay
        if name_mappings:
            for df_name, df_name_mapping in name_mappings.items():
                if df_name in pipeline.name_mappings:
                    pipeline.name_mappings[df_name].update(name_mappings[df_name])
                else:
                    pipeline.name_mappings[df_name] = name_mappings[df_name]

        # At this stage, ensure that the dataframes conforms to their metadata
        metadata = OrderedDict()
        for name, df in dataframes.items():
            try:
                df.validate_metadata()
            except ValueError as e:
                raise RuntimeError(
                    f"{self.__class__.__name__}.prepare: specification of the provided AnnotatedDataFrame '{name}'"
                    f" does not match its data. If you modified the frame after construction ensure"
                    f" consistency between 'data' and 'metadata' by using "
                    f" AnnotatedDataFrame.validate_metadata()"
                    f" -- {e}"
                ) from e

            # The pipeline will define a name mapping (only for items in its interface)
            if name in pipeline.name_mappings:
                for k, v in pipeline.name_mappings[name].items():
                    for node in pipeline.processing_graph.nodes():
                        pipeline_element = node.transformer

                        for dataset, mapping in pipeline_element.name_mappings.items():
                            for from_val, to_val in mapping.items():
                                if to_val == k:
                                    pipeline_element.name_mappings[dataset][from_val] = v

            metadata[name] = df.metadata

        validation_result = pipeline.validate(processing_graph=pipeline.processing_graph,
                                          metadata=metadata)

        pipeline.processing_graph = validation_result["processing_graph"]
        pipeline._output_specs: List[DataSpecification] = validation_result["output_spec"]

        pipeline.is_ready = True
        return pipeline

    def transform(self,
                  df: AnnotatedDataFrame,
                  name_mappings: dict[str, dict[str,any]] | None = None,
                  verbose: bool = False,
                  **kwargs) -> AnnotatedDataFrame:
        """
        Apply pipeline on given annotated dataframe

        .. note::
            If any filters are applied by the pipeline, these are executed at the end of the
            transformation.

        :param df: The input dataframe
        :returns: The transformed dataframe
        """
        dataframes = { DAMAST_DEFAULT_DATASOURCE: df }
        for x in self.processing_graph.get_joins():
            if x.name not in kwargs:
                raise RuntimeError("DataProcessingPipeline.transform: "
                                   "missing data source argument"
                                   f" for join '{x}' - found {kwargs.keys()}")
            dataframes[x.name] = kwargs[x.name]

        for label, dataframe in dataframes.items():
            if not isinstance(df, AnnotatedDataFrame):
                raise TypeError(
                    f"{self.__class__.__name__}.transform"
                    f" expected an annotated dataframe for processing ({label=}), "
                    f"got {type(df)}"
                )

        pipeline = self.prepare(name_mappings=name_mappings, **dataframes)

        for name, df in dataframes.items():
            if df.is_empty():
                raise RuntimeError(
                    f"{self.__class__.__name__}.transform: there is no data available to transform using datasource '{name}'"
            )

        if pipeline._inplace_transformation:
            in_dataframes = dataframes
        else:
            in_dataframes = {x: copy.deepcopy(y) for x,y in dataframes.items()}

        adf = pipeline._run(in_dataframes, verbose=verbose)
        assert isinstance(adf, AnnotatedDataFrame)

        adf.validate_metadata(validation_mode=damast.core.ValidationMode.UPDATE_METADATA)
        return adf

    def _run(self,
             dataframes: dict[str, AnnotatedDataFrame],
             verbose: bool = True) -> AnnotatedDataFrame:
        iterator = enumerate(self.processing_graph.nodes(), start=1)
        if not verbose:
            iterator = tqdm(iterator, desc="Step ", total=len(self.processing_graph))

        for idx, node in iterator:
            # ensure clean state
            if node.result:
                self.processing_graph.clear_state()

            try:
                logger.info("#{idx} run {node}")
                if node.name in dataframes and node.is_datasource():
                    node.result = node.transformer.fit_transform(dataframes[node.name])
                    AnnotatedDataFrame.ensure_type(node.result)
                else:
                    node.result = self.processing_graph.execute(node)

                if verbose:
                    print(f"Preview step #{idx} {node} (1 row)")
                    print(f"{node.result.head(1).collect()}")
            except Exception as e:
                msg = ''.join(tc.format_exception(e)[-2:])
                for slot, df in self.processing_graph.get_current_inputs(node).items():
                    msg += "{slot=} "
                    msg += "     {df.head(1).collect)}"

                if 'DAMAST_INTERACTIVE' in os.environ:
                    if str(os.environ['DAMAST_INTERACTIVE']).lower() == "true":
                        breakpoint()

                raise RuntimeError(f"Step #{idx} in pipeline ({node}) failed: name_mappings: {node.transformer.name_mappings}\n\
                        {msg}")
        return node.result


    def on_transform_start(self,
                           step: PipelineElement,
                           adf: AnnotatedDataFrame):
        """
        Default implementation of the on_transform_start callback.

        This output is a message in INFO log level
        """
        if hasattr(step, "transform_start"):
            return

        logger.info(f"[transform] start: {step.__class__.__name__} - {step.name_mappings}")
        start_time = datetime.now(timezone.utc)
        setattr(step, "transform_start", start_time)

        step_name = self.processing_graph[step.uuid].name
        self._processing_stats[step_name] = {
            "input_dataframe_length": adf.shape[0],
            "start_time": start_time
        }

    def on_transform_end(self,
                         step: PipelineElement,
                         adf: AnnotatedDataFrame):
        """
        Default implementation of the on_transform_end callback.

        This outputs a message in INFO log level
        """
        if not hasattr(step, "transform_start"):
            return

        start = getattr(step, "transform_start")
        end_time = datetime.now(timezone.utc)
        delta = (end_time - start).total_seconds()
        delattr(step, "transform_start")
        logger.info(f"[transform] end: {step.__class__.__name__} - {step.name_mappings}: "
                  f"{delta} seconds, {adf.shape[0]} remaining rows)")

        step_name = self.processing_graph[step.uuid].name
        self._processing_stats[step_name].update({
            "processing_time_in_s": delta,
            "output_dataframe_length": adf.shape[0],
            "end_time": end_time
        })

    def __repr__(self) -> str:
        """
        Create the representation string for this :class:`DataProcessingPipeline`.

        :returns: String representation
        """
        return self.to_str()

    def to_str(self, indent_level: int = 0) -> str:
        return self.processing_graph.to_str(indent_level=indent_level)

    def __deepcopy__(self, memo) -> DataProcessingPipeline:
        pipeline = copy.copy(self)

        pipeline._name_mappings = copy.deepcopy(self._name_mappings)
        pipeline._processing_stats = copy.deepcopy(self._processing_stats)
        pipeline.processing_graph = copy.deepcopy(self.processing_graph)
        pipeline._output_specs = [copy.deepcopy(x) for x in self._output_specs]
        return pipeline

def polar_data_type_constructor(loader, tag_suffix, node):
    module_whitelist = ["polars.datatypes.classes"]
    for m in module_whitelist:
        if tag_suffix.startswith(m):
            module = importlib.import_module(m)
            object_name = tag_suffix.replace(m + ".","")
            return getattr(module, object_name)

    raise ValueError(f"damast.core.dataprocessing: no constructor registered for {tag_suffix=}")

yaml.SafeLoader.add_multi_constructor("tag:yaml.org,2002:python/name:", polar_data_type_constructor)



import functools
import inspect
from collections import OrderedDict

from damast.core.dataframe import AnnotatedDataFrame

from .constants import (
    DAMAST_DEFAULT_DATASOURCE,
    DECORATED_ARTIFACT_SPECS,
    DECORATED_DESCRIPTION,
    DECORATED_INPUT_SPECS,
    DECORATED_OUTPUT_SPECS,
    )
from .metadata import ArtifactSpecification, DataSpecification
from .transformations import PipelineElement


def _get_dataframe(*args, **kwargs) -> AnnotatedDataFrame:
    """
    Extract the dataframe from positional or keyword arguments
    :param datasource: the expected name of the datasource
    :param args: positional arguments
    :param kwargs: keyword arguments
    :return: The annotated data frame
    :raise KeyError: if a positional argument does not exist and keyword :code:`df` is missing
    :raise TypeError: if the kwargs 'df' is not an AnnotatedDataFrame
    """
    _df: AnnotatedDataFrame

    arguments = kwargs.copy()
    if 'datasource_argname' in arguments:
        argname = arguments['datasource_argname']
    else:
        argname = DAMAST_DEFAULT_DATASOURCE

    arg_index = 0
    for parameter in inspect.signature(args[0].transform).parameters:
        if parameter not in kwargs:
            arg_index += 1
            arguments[parameter] = args[arg_index]

    datasource = arguments[argname]
    if not isinstance(datasource, AnnotatedDataFrame):
        raise TypeError(f"Argument {argname} is not an AnnotatedDataFrame")

    return datasource

def describe(description: str):
    """
    Specify the description for the transformation for the decorated function.

    The decorated function must return :class:`damast.core.AnnotatedDataFrame`.

    :param description: description of the action
    """

    def decorator(func):
        setattr(func, DECORATED_DESCRIPTION, description)
        return func

    return decorator


def input(requirements: dict[str, any], label: str | None = None):
    """
    Specify the input for the decorated function.

    The decorated function must return :class:`damast.core.AnnotatedDataFrame`.

    :param requirements: List of input requirements
    """

    if label is None:
        label = DAMAST_DEFAULT_DATASOURCE

    required_input_specs = DataSpecification.from_requirements(
        requirements=requirements
    )

    def decorator(func):
        if not hasattr(func, DECORATED_INPUT_SPECS):
            setattr(func, DECORATED_INPUT_SPECS, OrderedDict())

        input_specs = getattr(func, DECORATED_INPUT_SPECS)
        if label in input_specs:
            raise KeyError(f"input for '{label}' has already been specified")

        parameters = list(inspect.signature(func).parameters.keys())
        if label not in parameters:
            for x,_ in parameters:
                if x == label:
                    break

                if x not in input_specs:
                    raise KeyError(f"input for '{x}' unknown, but needs to be before '{label}' - please validate decorator order")

        input_specs[label] = required_input_specs

        @functools.wraps(func)
        def check(*args, **kwargs):
            _df: AnnotatedDataFrame = _get_dataframe(*args, datasource_argname=label, **kwargs)
            # Ensure that name mapping are applied correctly
            assert isinstance(args[0], PipelineElement)

            pipeline_element = args[0]
            fulfillment = _df.get_fulfillment(expected_specs=pipeline_element.input_specs[label])
            if fulfillment.is_met():
                if hasattr(pipeline_element, "parent_pipeline"):
                    # if this is the last datasource parameter, this is also the last input decorators
                    # so the transform can start
                    if label == parameters[-1]:
                        getattr(pipeline_element, "parent_pipeline").on_transform_start(pipeline_element, adf=_df)
                return func(*args, **kwargs)

            raise RuntimeError(
                "Input requirements are not fulfilled:" f" -- {fulfillment}"
            )

        return check

    return decorator


def output(requirements: dict[str, any]):
    """
    Specify the output for the decorated function.

    The decorated function must return :class:`damast.core.AnnotatedDataFrame`.

    :param requirements: List of input requirements
    """
    required_output_specs = DataSpecification.from_requirements(
        requirements=requirements
    )

    def decorator(func):
        return_type = inspect.signature(func).return_annotation
        if return_type != AnnotatedDataFrame and return_type != 'AnnotatedDataFrame':
            raise RuntimeError(
                f"output: decorator requires 'AnnotatedDataFrame' to be returned by function - but was '{return_type}'"
            )

        setattr(func, DECORATED_OUTPUT_SPECS, required_output_specs)

        @functools.wraps(func)
        def check(*args, **kwargs) -> AnnotatedDataFrame:
            setattr(func, DECORATED_OUTPUT_SPECS, required_output_specs)

            _df: AnnotatedDataFrame = _get_dataframe(*args, **kwargs)
            input_columns = list(_df.column_names)

            adf: AnnotatedDataFrame = func(*args, **kwargs)
            if adf is None:
                raise RuntimeError(
                    f"output: decorated function {func} must return 'AnnotatedDataFrame',"
                    f" but was 'None'"
                )

            if not isinstance(adf, AnnotatedDataFrame):
                raise RuntimeError(
                    f"output: decorated function {func} must return 'AnnotatedDataFrame', but was '"
                    f"{type(adf)}"
                )

            for c in input_columns:
                if c not in adf.column_names:
                    raise RuntimeError(
                        f"output: column '{c}' was removed by decorated function."
                        f" Only adding of columns is permitted."
                    )

            # Ensure that name mapping are applied correctly
            assert isinstance(args[0], PipelineElement)
            pipeline_element = args[0]

            parent_pipeline = None
            if hasattr(pipeline_element, "parent_pipeline"):
                parent_pipeline = getattr(pipeline_element, "parent_pipeline")

            try:
                # Ensure that metadata is up to date with the dataframe
                adf.update(expectations=pipeline_element.output_specs)
            except RuntimeError as e:
                txt = f"Failed to update metadata in pipeline element: {pipeline_element}"
                if parent_pipeline:
                    txt += f" in pipeline '{parent_pipeline.name}' ({type(parent_pipeline)})"
                txt += f" -- {e}"

                raise RuntimeError(txt) from e

            if parent_pipeline:
                parent_pipeline.on_transform_end(pipeline_element, adf=adf)

            return adf

        return check

    return decorator


def artifacts(requirements: dict[str, any]):
    """
    Specify the output for the decorated function.

    The decorated function must return :class:`damast.core.AnnotatedDataFrame`.

    :param requirements: List of input requirements
    """
    required_artifact_specs = ArtifactSpecification(requirements=requirements)

    def decorator(func):
        return_type = inspect.signature(func).return_annotation
        if return_type != AnnotatedDataFrame:
            raise RuntimeError(
                "artifacts: decorator requires 'AnnotatedDataFrame' to be returned by function"
            )

        setattr(func, DECORATED_ARTIFACT_SPECS, required_artifact_specs)

        # When a pipeline does generate artifacts, then it might not provide any output, but serves only as
        # passthrough element. Hence, per default set an empty output spec if there is none.
        if not hasattr(func, DECORATED_OUTPUT_SPECS):
            setattr(
                func,
                DECORATED_OUTPUT_SPECS,
                DataSpecification.from_requirements(requirements={}),
            )

        @functools.wraps(func)
        def check(*args, **kwargs) -> AnnotatedDataFrame:
            result = func(*args, **kwargs)

            if not isinstance(args[0], PipelineElement):
                raise RuntimeError(
                    f"{args[0].__class__.__name__} must inherit from PipelineElement"
                )

            # Validate the spec with respect to the existing parent pipeline
            instance = args[0]
            try:
                required_artifact_specs.validate(
                    base_dir=instance.parent_pipeline.base_dir
                )
            except RuntimeError as e:
                raise RuntimeError(
                    f"artifacts: {func} is expected to generate an artifact. "
                    f" Pipeline element ran as part of pipeline: '{instance.parent_pipeline.name}'"
                ) from e

            return result

        return check

    return decorator




import sys
import importlib

if sys.version_info < (3,11):
    import iso8601
    def fromisoformat(iso_string: str):
        return iso8601.parse_date(iso_string)
else:
    import datetime as dt
    def fromisoformat(iso_string: str):
        return dt.datetime.fromisoformat(iso_string)


def ensure_packages(pkgs: list[str], required_for: str, hint: str | None = None,
                    install: dict[str, str] = {}):
    requirements = []
    for pkg in pkgs:
        if pkg in install:
            requirements.append(install[pkg])
        else:
            requirements.append(pkg)

        msg = f"{required_for} requires to 'pip install {' '.join(requirements)}'"

        if hint is not None:
            msg += f", {hint}"

        if not importlib.util.find_spec(pkg):
            raise RuntimeError(msg)



import logging
import re
from argparse import ArgumentParser
from pathlib import Path

import polars as pl

from damast.cli.base import BaseParser
from damast.core.dataframe import AnnotatedDataFrame
from damast.utils.io import Archive

logger = logging.getLogger(__name__)

class DataInspectParser(BaseParser):
    """
    Argparser for inspecting AnnotatedDataFrame

    :param parser: The base parser
    """

    def __init__(self, parser: ArgumentParser):
        super().__init__(parser=parser)

        parser.description = "damast inspect - data inspection subcommand called"
        parser.add_argument("-f", "--files",
                            help="Files or patterns of the (annotated) data file that should be inspected (space separated)",
                            nargs="+",
                            type=str,
                            required=True
                            )

        parser.add_argument("--filter",
                help="Filter based on column data, e.g., mmsi==120123",
                action="append"
        )
        parser.add_argument("--head", type=int, default=10, help="First this number of rows, default is 10")
        parser.add_argument("--tail", type=int, default=10, help="Print number of rows from the end, default is 10")
        parser.add_argument("--column-count", type=int, default=10, help="Number of columns to show")

        parser.add_argument("--columns",
                help="Show/Select these columns",
                nargs="+",
                type=str,
                required=False
        )

    def expand_filter_arg(self, adf: AnnotatedDataFrame, arg: str):
        if arg in adf.column_names:
            return f"pl.col('{arg}')"

        m = re.match(r"datetime\((.*)\)", arg)
        if m:
            return f"damast.utils.fromisoformat({m.group(1)})"

        return arg

    def execute(self, args):
        super().execute(args)

        files_stats = self.get_files_stats(args.files)
        print(f"Loading dataframe ({files_stats.number_of_files} files) of total size: {files_stats.total_size} MB")

        try:
            with Archive(filenames=args.files) as input_files:
                files = [x for x in input_files if AnnotatedDataFrame.get_supported_format(Path(x).suffix)]
                if not files:
                    raise RuntimeError(f"Inspection is not supported for input files: {input_files=}")

                adf = AnnotatedDataFrame.from_files(files=files, metadata_required=False)

                if args.filter:
                    filter_values = ""
                    for filter_expression in args.filter:
                        m = re.match(r"([^!=<>]+)([!=><]+)([^!=<>]*)", filter_expression)
                        if m:
                            lhs = m.group(1).strip()
                            op = m.group(2).strip()
                            rhs = m.group(3).strip()

                            lhs = self.expand_filter_arg(adf, lhs)

                            new_filter = ""
                            if rhs in ["null", "None"]:
                                if op == "==":
                                    new_filter = f"{lhs}.is_null()"
                                elif op == "!=":
                                    new_filter = f"{lhs}.is_not_null()"
                                else:
                                    logger.warning("Filter expression invalid: operator must be either '==' or '!='")
                                    continue
                            else:
                                rhs = self.expand_filter_arg(adf, rhs)
                                new_filter = f"{lhs} {op} {rhs}"

                            print(f"   .filter({new_filter})")
                            filter_values += f".filter({new_filter})"
                        else:
                            logger.warning(f"Filter expression invalid: {filter_expression}")

                    adf._dataframe = eval(f"adf._dataframe{filter_values}")
                    adf._metadata = AnnotatedDataFrame.infer_annotation(adf._dataframe)

                print(adf.metadata.to_str(columns=args.columns))
                print(f"\n\nFirst {args.head} and last {args.tail} rows:")
                df = adf._dataframe
                if args.columns:
                    df = df.select(args.columns)

                with pl.Config(tbl_rows=args.head, tbl_cols=args.column_count):
                    print(df.head(n=args.head).collect())
                with pl.Config(tbl_rows=args.tail, tbl_cols=args.column_count):
                    print(df.tail(n=args.tail).collect())

        except RuntimeError as e:
            if re.search(r"metadata is missing", str(e)) is not None:
                print(e)
            else:
                raise



import numpy as np
import polars

from damast.core.data_description import MinMax
from damast.core.dataframe import AnnotatedDataFrame
from damast.core.decorators import describe, input, output
from damast.core.transformations import PipelineElement


class CycleTransformer(PipelineElement):
    def __init__(self, n: int):
        self.n = n

    @describe("Cycle transformation")
    @input({'x': {}})
    @output({
        "{{x}}_x": {"value_range": MinMax(-1.0, 1.0)},
        "{{x}}_y": {"value_range": MinMax(-1.0, 1.0)},
    })
    def transform(self, df: AnnotatedDataFrame) -> AnnotatedDataFrame:
        feature = self.get_name('x')
        df._dataframe = df._dataframe.with_columns(
                (np.cos(polars.col(feature)*2*np.pi) / self.n).alias(f"{feature}_x"),
                (np.cos(polars.col(feature)*2*np.pi) / self.n).alias(f"{feature}_y")
        )
        return df
